/*@
theory uses Base {
  int: TYPE = Base.int;
  intArray: TYPE = Base.IntArray;
  nullArray: intArray = Base.nullIntArray;
  isSorted: PREDICATE(intArray, int, int) =
    PRED(a: intArray, from: int, to: int): 
      FORALL(i: INT): from <= i AND i < to-1 => a.value[i] <= a.value[i+1];
  arePermutation: PREDICATE(intArray, intArray, int) =
    PRED(a: intArray, b: intArray, n: int):
      EXISTS(p: ARRAY NAT OF NAT):
        (FORALL(j: NAT): j < n => (EXISTS(i: NAT): i < n AND p[i]=j)) AND
        (FORALL(i: NAT): i < n => p[i] < n AND a.value[i] = b.value[p[i]]);
  areIdentical: PREDICATE(intArray, intArray, int, int) =
    PRED(a: intArray, b: intArray, from: int, to: int):
       FORALL(i: int): from <= i AND i < to => a.value[i] = b.value[i];
}
@*/
class Arrays1
{ 
  public static void insertionSort(int[] a) /*@
    requires NOT (VAR a).null;
    assignable a;
    ensures NOT (VAR a).null AND
       (LET n = (OLD a).length IN
       n = (VAR a).length AND
       isSorted(VAR a, 0, n) AND
       arePermutation(VAR a, OLD a, n));
  @*/
  {
    int n = a.length; 
    for (int i = 1; i < n; i++) /*@
    invariant
       NOT (VAR a).null AND NOT (OLD a).null AND
       VAR n = (OLD a).length AND VAR n = (VAR a).length AND
       1 <= VAR i AND (VAR n > 0 => VAR i <= VAR n) AND
       areIdentical(VAR a, OLD a, VAR i, VAR n) AND
       isSorted(VAR a, 0, VAR i) AND 
       arePermutation(VAR a, OLD a, VAR n);
    decreases VAR n - VAR i + 1;
    @*/
      insert(a, i);
  }
  
  private static void insert(int[] a, int i) /*@
    requires NOT (VAR a).null AND (VAR a).length > 1 AND
      1 <= VAR i AND VAR i < (VAR a).length AND
      isSorted(VAR a, 0, VAR i);
    assignable a;
    ensures NOT (VAR a).null AND
      (VAR a).length = (OLD a).length AND
      isSorted(VAR a, 0, VAR i+1) AND
      arePermutation(VAR a, OLD a, VAR i+1) AND
      areIdentical(VAR a, OLD a, VAR i+1, (VAR a).length);
   @*/
  {
    int x = a[i];
    int j = i-1;
    while (j >= 0 && a[j] > x) /*@
      invariant
        NOT (VAR a).null AND NOT (OLD a).null AND
        (LET n = (OLD a).length IN
        n = (VAR a).length AND
        1 <= VAR i AND VAR i < n AND
        VAR x = (OLD a).value[VAR i] AND
        -1 <= VAR j AND VAR j <= VAR i - 1 AND
        areIdentical(VAR a, OLD a, 1+VAR i, n) AND
        areIdentical(VAR a, OLD a, 0, 1+VAR j) AND
        (FORALL(k:int): VAR j < k AND k < VAR i =>
          (VAR a).value[k+1] = (OLD a).value[k] AND
          (OLD a).value[k] > VAR x));
      decreases VAR j + 1;
      @*/ 
    {
      a[j+1] = a[j];
      j = j-1;
    }
    a[j+1] = x;
  }
}








